/*
 * 	dcpcgi.c
 * 	
 * 	04/24/2000 Matthew T. Kromer
 *
 * 	Copyright (c) 2000 Digital Creations, Inc. 
 *
 * 	$Id: dcpcgi.c,v 1.1 2000/06/05 19:18:50 matt Exp matt $
 *
 # Zope Public License (ZPL) Version 1.0
 # -------------------------------------
 # 
 # Copyright (c) Digital Creations.  All rights reserved.
 # 
 # This license has been certified as Open Source(tm).
 # 
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are
 # met:
 # 
 # 1. Redistributions in source code must retain the above copyright
 #    notice, this list of conditions, and the following disclaimer.
 # 
 # 2. Redistributions in binary form must reproduce the above copyright
 #    notice, this list of conditions, and the following disclaimer in
 #    the documentation and/or other materials provided with the
 #    distribution.
 # 
 # 3. Digital Creations requests that attribution be given to Zope
 #    in any manner possible. Zope includes a "Powered by Zope"
 #    button that is installed by default. While it is not a license
 #    violation to remove this button, it is requested that the
 #    attribution remain. A significant investment has been put
 #    into Zope, and this effort will continue if the Zope community
 #    continues to grow. This is one way to assure that growth.
 # 
 # 4. All advertising materials and documentation mentioning
 #    features derived from or use of this software must display
 #    the following acknowledgement:
 # 
 #      "This product includes software developed by Digital Creations
 #      for use in the Z Object Publishing Environment
 #      (http://www.zope.org/)."
 # 
 #    In the event that the product being advertised includes an
 #    intact Zope distribution (with copyright and license included)
 #    then this clause is waived.
 # 
 # 5. Names associated with Zope or Digital Creations must not be used to
 #    endorse or promote products derived from this software without
 #    prior written permission from Digital Creations.
 # 
 # 6. Modified redistributions of any form whatsoever must retain
 #    the following acknowledgment:
 # 
 #      "This product includes software developed by Digital Creations
 #      for use in the Z Object Publishing Environment
 #      (http://www.zope.org/)."
 # 
 #    Intact (re-)distributions of any official Zope release do not
 #    require an external acknowledgement.
 # 
 # 7. Modifications are encouraged but must be packaged separately as
 #    patches to official Zope releases.  Distributions that do not
 #    clearly separate the patches from the original work must be clearly
 #    labeled as unofficial distributions.  Modifications which do not
 #    carry the name Zope may be packaged in any form, as long as they
 #    conform to all of the clauses above.
 # 
 # 
 # Disclaimer
 # 
 #   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
 #   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 #   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 #   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
 #   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 #   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 #   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 #   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 #   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 #   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 #   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 #   SUCH DAMAGE.
 # 
 # 
 # This software consists of contributions made by Digital Creations and
 # many individuals on behalf of Digital Creations.  Specific
 # attributions are listed in the accompanying credits file.
 *************************************************************************
 *
 * Usage:
 *
 * In the obj.conf file, in the Init section, add:
 *
 * Init fn="load-modules" shlib="/usr/netscape/server4/plugins/lib/dcpcgi.so" funcs="dcpcgi-send,dcpcgi-path"
 *
 * (adjusting the shlib path to wherever the dcpcgi.so file really lives)
 *
 *
 * Then, include the following in the "default" object
 *
 * NameTrans fn="pfx2dir" from="/Zope" dir="/Zope" name="Zope"
 *
 *
 * Then create the following Object:
 *
 * <Object name="Zope">
 * PathCheck fn="dcpcgi-path" base="/Zope"
 * Service fn="dcpcgi-send" pcgi-socket="/home/matt/zope/var/pcgi.soc"
 * </Object>
 *
 * Where the pcgi-socket variable is wherever the pcgi socket for your Zope
 * instance lives.
 *
 *************************************************************************
 *
 * TODO:
 *
 *	Fix up the NameTrans/PathCheck stuff so that a simple assign-name
 * 	will work (e.g. assign a style to a directory) and make it smart enough
 * 	to know where it is so the base= parameter is not required to do
 * 	PATH_INFO reformulations.
 *
 *	Handle large file inputs by interleaving network reads with pcgi
 *	writes rather than buffering the entire input
 * 
 */

#ifdef XP_WIN32
#	define NSAPI_PUBLIC __declspec(dllexport)
#else /* !XP_WIN32 */
#	define NSAPI_PUBLIC
#endif

#include "nsapi.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>


/*	
**
** The following code is probably unix only right now
**
*/


#ifndef BUFFER_ALLOCATION
#define BUFFER_ALLOCATION 16384
#endif

typedef struct {
	char *buffer;
	unsigned int size;
	unsigned int used;
	char *read;
	char *write;
} Buffer;

typedef struct {
	int inlen;
	int errlen;
	int state;
} PrintState;

#define NSAPIARGS pblock *param, Session *sn, Request *rq
#define NSAPIPASS param, sn, rq

static char *DCCopyright="$Id: dcpcgi.c,v 1.1 2000/06/05 19:18:50 matt Exp matt $\nCopyright (c) 2000 Digital Creations.";

/* 
** bufferFree
**
** release a buffer
**
*/

void bufferFree(Buffer *buf) {
	if (buf == NULL) return;

	FREE(buf->buffer);
	FREE(buf);
}

/* 
** openSocket(filename)
**
** filename is the AF_Unix socket name, or the hostname:port to connect to
** the type is AF_UNIX or AF_INET depending on the presence of the colon
*/

int openSocket(char *filename, NSAPIARGS) {
	
	int fd;
	int type = AF_INET;
	struct protoent *pent; 
	char *hostname;
	unsigned long one = 1;
	hostname = filename;
	while (*filename  && *filename != ':') filename++;
	if (*filename != ':') type = AF_UNIX;

	if ((pent = getprotobyname("tcp")) == NULL) {
		log_error(LOG_FAILURE, "dcpcgi-openSocket", sn, rq, 
			"getprotobyname(tcp) failed: %s", strerror(errno));
		return -1;
	}

	if (type == AF_UNIX) {
		struct sockaddr_un saddr;
		saddr.sun_family = AF_UNIX;
		strncpy(saddr.sun_path, hostname, sizeof(saddr.sun_path)-1);

		if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
			log_error(LOG_FAILURE, "dcpcgi-openSocket", sn, rq, 
				"socket(AF_UNIX) failed: %s", strerror(errno));
			return -1;
		}

		if (connect(fd, &saddr, sizeof(saddr)) < 0) {
			log_error(LOG_FAILURE, "dcpcgi-openSocket", sn, rq, 
				"connect failed: %s", strerror(errno));
			return -1;
		}
		
	} else if (type == AF_INET) {
		struct sockaddr_in saddr;
		unsigned short port;
		struct hostent *hent;

		*filename = '\0';
		filename++;
		port = atoi(filename);

		if ((hent = gethostbyname(hostname)) == NULL) {
			log_error(LOG_FAILURE, "dcpcgi-openSocket", sn, rq, 
				"gethostbyname failed: %s", strerror(errno));
			return -1;
		}

		saddr.sin_family = AF_INET;
		saddr.sin_port = htons(port);
		memcpy((char *) saddr.sin_addr.s_addr,
			hent->h_addr, hent->h_length);

		if ((fd = socket(AF_INET, SOCK_STREAM, pent->p_proto)) < 0) {
			log_error(LOG_FAILURE, "dcpcgi-openSocket", sn, rq, 
				"socket(AF_INET) failed: %s", strerror(errno));
			return -1;
		}

		if (connect(fd, &saddr, sizeof(saddr)) < 0) {
			log_error(LOG_FAILURE, "dcpcgi-openSocket", sn, rq, 
				"connect failed: %s", strerror(errno));
			return -1;
		}

	} 

	return fd;
}

/*
** BufferWrite
**
*/

Buffer *bufferWrite(Buffer *buf, char *string, int len, NSAPIARGS) {
	char *debug=pblock_findval("debug",param);

	if (buf == NULL) {
		if ((buf = (Buffer *) CALLOC(sizeof(Buffer))) == NULL) {
			log_error(LOG_FAILURE, "dcpcgi-bufferWrite", sn, rq, 
				"CALLOC failed");
			return NULL;
		}
		buf->buffer = (char *) CALLOC(BUFFER_ALLOCATION);
		buf->size = BUFFER_ALLOCATION;
		buf->read = buf->buffer;
		buf->write = buf->buffer;
	}

	if (debug != NULL && strcmp(debug,"all") == 0) {
		log_error(LOG_INFORM, "dcpcgi-bufferWrite", sn, rq,
		"Writing %d characters to buffer: %s", len, string);
	}

	if (buf->used + len > buf->size -1) {
		unsigned int readc;
		unsigned int writec;

		readc = buf->read - buf->buffer;
		writec = buf->write - buf->buffer;

		if (debug != NULL) 
			log_error(LOG_INFORM, "dcpcgi-bufferWrite", sn, rq,
			"Buffer %08x growing by %d bytes, head @ %08x: %10.10s",
			(unsigned long) buf, BUFFER_ALLOCATION,
			(unsigned long) buf->read, buf->read);

		if ((buf->buffer = (char *) REALLOC(buf->buffer, buf->size +
			BUFFER_ALLOCATION)) == NULL) {
			log_error(LOG_FAILURE, "dcpcgi-bufferWrite", sn, rq, 
				"REALLOC failed");
			return NULL;
		}

		buf->size += BUFFER_ALLOCATION;

		buf->read = buf->buffer + readc;	/* Pointers moved */
		buf->write = buf->buffer + writec;

		if (debug != NULL) 
			log_error(LOG_INFORM, "dcpcgi-bufferWrite", sn, rq,
			"Buffer %08x grew by %d bytes, head @ %08x: %10.10s",
			(unsigned long) buf, BUFFER_ALLOCATION, 
			(unsigned long) buf->read, buf->read);
	}

	memcpy(buf->write,string,len);
	buf->used += len;
	buf->write += len;

	if (debug != NULL) 
		log_error(LOG_INFORM, "dcpcgi-bufferWrite", sn, rq,
		"Buffer %08x wrote %d chars, head at %08x = %10.10s, tail at %08x",
		(unsigned long) buf, len, (unsigned long) buf->read, buf->read,
		(unsigned long) buf->write);
			

	return buf;
}

/*
** bufferXmit
** 
*/

int bufferXmit(Buffer *buf, int fd, NSAPIARGS) {

	int remain;
	int rc;
	int count = 0;
	char *debug = pblock_findval("debug", param);
	remain = buf->used - (buf->read - buf->buffer);

	if (debug != NULL) {
		log_error(LOG_INFORM, "dcpcgi-bufferxmit", sn, rq,
			"writing %d bytes to socket %d", remain, fd);
	}

	while (remain) {
		rc = write(fd, buf->read, remain);
		if (rc < 0) {
			if (errno != EAGAIN) {
				log_error(LOG_FAILURE, "dcpcgi-bufferXmit",
					sn, rq, "write failed: %s",
					strerror(errno));
				return REQ_ABORTED;
			}
		} else {
			remain -= rc;
			buf->read += rc;
			count += rc;
		}
	}

	return count;
}

/*
** bufferReset
**
*/

void bufferReset(Buffer *buf, NSAPIARGS) {
	char *debug = pblock_findval("debug",param);

	buf->read = buf->buffer;
	buf->write = buf->buffer;
	buf->used = 0;

	if (debug != NULL) 
	    log_error(LOG_INFORM, "dcpcgi-bufferReset", sn, rq, 
	    	"buffer reset at %08x", (unsigned long) buf);
}

/*
** bufferRead (read from the network)
**
*/

int bufferRead(Buffer *buf, int fd, NSAPIARGS) {
	char tbuf[8192];
	int rc;
	int count = 0;
	char *debug = pblock_findval("debug", param);
	fd_set readfds;
	fd_set writefds;
	fd_set exceptfds;
	struct timeval timeout;

	while (1) {
		if (debug != NULL) {
			log_error(LOG_INFORM, "dcpcgi-bufferxmit", sn, rq,
				"reading from socket %d", fd);
		}

		FD_ZERO(&readfds);
		FD_ZERO(&writefds);
		FD_ZERO(&exceptfds);

		FD_SET(fd, &readfds);
		FD_SET(fd, &exceptfds);

		timeout.tv_sec = 30;
		timeout.tv_usec = 0;

#if 1
		rc = net_select(fd+1, &readfds, &writefds, &exceptfds, &timeout);

		if (rc < 1) return count;
#endif

		rc = read(fd, tbuf, sizeof(tbuf));

		if (debug != NULL) {
			log_error(LOG_INFORM, "dcpcgi-bufferxmit", sn, rq,
				"rc = %d from read socket %d", rc, fd);
		}
		if (rc < 0) {
			if (errno != EAGAIN) {
				log_error(LOG_FAILURE, "dcpcgi-bufferRead",
					sn, rq, "read failed: %s",
					strerror(errno));
				return REQ_ABORTED;
			}
		} else if (rc > 0) {
			count += rc;
			bufferWrite(buf, tbuf, rc, NSAPIPASS);
		} else break;
	}
	return count;
}

/*
** nsbufferRead (read from the client)
**
*/

Buffer *nsbufferRead(Buffer *buf, int size, NSAPIARGS) {
	char tbuf[8192];
	int rc;
	int count = 0;

	while (size == 0 || count < size) {
		rc = net_read(sn->csd, tbuf, sizeof(tbuf), 1);
		if (rc < 0) {
			log_error(LOG_FAILURE, "dcpcgi-nsbufferRead",
				sn, rq, "read failed: %s",
				strerror(errno));
			return NULL;
		} else if (rc > 0) {
			count += rc;	/* XXX count could exceed size here */
			buf = bufferWrite(buf, tbuf, rc, NSAPIPASS);
		} else break;
	}
	return buf;
}

/*
** bufferPrint (send the buffer to the client)
**
**
*/

void bufferPrint(Buffer *buf, PrintState *state, NSAPIARGS) {
	int remain;
	char *proto;
	char *debug = pblock_findval("debug",param);

	remain = buf->used - (buf->read - buf->buffer);

	if (state->state == 0) {
		if (remain > 10) {
			state->inlen = atoi(buf->read);
			buf->read += 10;
			remain -= 10;

			if (debug != NULL)
			    log_error(LOG_INFORM,"dcpcgi-bufferPrint",
				sn, rq, "stdin length is %d", state->inlen);
			if (state->inlen == 0) {
			    log_error(LOG_WARN,"dcpcgi-bufferPrint",
				sn, rq, "stdin length is 0! Buf head "
				"@ %08x: %10.10s", (buf->read - 10),
				(char *) (buf->read - 10));
			}

			/* Flush out that Status: header */
#if 1
			if ((proto = pblock_findval("protocol",
				rq->reqpb)) != NULL && remain >= 7) {
				int l = strlen(proto);
				if (l < 17) {
					buf->read =  buf->read + 7 - l;
					remain = remain - 7 + l;
					state->inlen = state->inlen - 7 + l;
					memcpy(buf->read, proto, l);
				}
			}
#endif
		}
		state->state = 1;
	}
	
	while (remain) {
		if (state->state == 1) {
			if (state->inlen >= remain) {
				net_write(sn->csd, buf->read, remain);
				state->inlen -= remain;
				remain = 0;
			} else {
				if (state->inlen == 0 && remain >= 10) {
					state->errlen = atoi(buf->read);
					buf->read += 10;
					remain -= 10;
					state->state = 2;
				} else {
					net_write(sn->csd, buf->read,
						state->inlen);
					buf->read += state->inlen;
					state->inlen = 0;
					remain -= state->inlen;
				}
			}
		} else {
			/* Stderr -- toss it */
			remain = 0;
		}
	}

	bufferReset(buf, NSAPIPASS);
}

/*
** nsapiEnvironment
** 
** Generate the CGI environments from the available NSAPI data
*/

Buffer *nsapiEnvironment(Buffer *buffer, NSAPIARGS) {
	char buf[8192];
	char *value;
	int len = 0;
	int accum = 0;
	char *lenat;
	
	sprintf (buf,"%010ld", 0);
	/* We'll come back and correct the size later */
	buffer = bufferWrite(buffer, buf, strlen(buf), NSAPIPASS); 
	lenat = buffer->write - strlen(buf);
	
	if ((value = pblock_findval("authorization", rq->headers)) != NULL) {
		sprintf(buf,"HTTP_CGI_AUTHORIZATION=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
	}

	if ((value = pblock_findval("auth-type", rq->vars)) != NULL) {
		sprintf(buf,"AUTH_TYPE=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
	}

	if ((value = pblock_findval("auth-user", rq->vars)) != NULL) {
		sprintf(buf,"AUTH_USER=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);

		sprintf(buf,"REMOTE_USER=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
	}

	if ((value = pblock_findval("content-length", rq->srvhdrs)) != NULL) {
		sprintf(buf,"CONTENT_LENGTH=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
	}

	if ((value = pblock_findval("content-type", rq->srvhdrs)) != NULL) {
		sprintf(buf,"CONTENT_TYPE=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
	}
	
	sprintf(buf,"GATEWAY_INTERFACE=CGI/1.1");
	len = strlen(buf) + 1;
	accum += len;
	buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
	
	if ((value = pblock_findval("path-info", rq->vars)) != NULL) {
		sprintf(buf,"PATH_INFO=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
	}

	if ((value = pblock_findval("path-translated", rq->vars)) != NULL) {
		sprintf(buf,"PATH_TRANSLATED=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
	}

	if ((value = pblock_findval("query", rq->reqpb)) != NULL) {
		sprintf(buf,"QUERY_STRING=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);

		sprintf(buf,"QUERY=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
	}

	if ((value = pblock_findval("ip", sn->client)) != NULL) {
		sprintf(buf,"REMOTE_ADDR=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
		
		sprintf(buf,"REMOTE_HOST=%s", session_dns(sn) ?
		      session_dns(sn) : value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
	}

	if ((value = pblock_findval("method", rq->reqpb)) != NULL) {
		sprintf(buf,"REQUEST_METHOD=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
	}

	if ((value = pblock_findval("uri", rq->reqpb)) != NULL) {
		char *c;
		char v;
		c = value;
		v = *c;	/* Save off the first character (a / most likely) */
		while (*c) c++;
		c--;
		while (*c == '/') {
			*c = '\0';
			c--;
		}
		*value = v; /* Restore the first character */
		sprintf(buf,"SCRIPT_NAME=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
	}
	
	sprintf(buf,"SERVER_NAME=%s", util_hostname());
	len = strlen(buf) + 1;
	accum += len;
	buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);

	sprintf(buf,"SERVER_PORT=%d", conf_getglobals()->Vport);
	len = strlen(buf) + 1;
	accum += len;
	buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);

#if 0
	sprintf(buf,"SERVER_SOFTWARE=%s", system_version);
	len = strlen(buf) + 1;
	accum += len;
	buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
#endif

	if ((value = pblock_findval("protocol", rq->reqpb)) != NULL) {
		sprintf(buf,"PROTOCOL=%s", value);
		len = strlen(buf) + 1;
		accum += len;
		buffer = bufferWrite(buffer, buf, strlen(buf)+1, NSAPIPASS);
	}
	
	sprintf(buf,"%010ld", accum);
	memcpy(lenat, buf, strlen(buf));

	return buffer;
}

/*
** main
**
*/

NSAPI_PUBLIC int dcpcgi_send(NSAPIARGS) {
	int	sock;
	Buffer *buffer = NULL;
	Buffer *inbuf = NULL;
	char *hostname;
	int envsize = 0;
	char *sockname;
	char buf[16];
	char *method;
	char *contentLength;
	int size;
	PrintState printstate;

	if ((sockname = pblock_findval("pcgi-socket", param)) == NULL) {
		log_error(LOG_MISCONFIG, "dcpcgi-send", sn, rq,
			"pcgi-socket not specified");
		return REQ_ABORTED;
	}


	sock = openSocket(sockname, NSAPIPASS);
	if (sock < 0) return REQ_ABORTED;	/* error condition */

	/* create the environment */
	buffer = nsapiEnvironment(buffer, NSAPIPASS);

	/* Get the type of method */
	method = pblock_findval("method", rq->reqpb);
	if (strncmp(method,"POST",5) == 0 || strncmp(method,"PUT",4) == 0) {
		contentLength = pblock_findval("content-length",rq->srvhdrs);
		size = atoi(contentLength);

		/* TODO:  For "large" size requests, print the stdin len
		** and begin cutting over buffers to the pcgi socket as
		** they are read; a short read will have to be handled by
		** the medusa server
		**
		** This means changing nsbufferRead to return after some
		** reasonable size input has been read or making it
		** call bufferXmit and bufferReset
		*/

		/* Capture stdin from the net */
		inbuf = nsbufferRead(inbuf, size, NSAPIPASS);
		if (inbuf == NULL) return REQ_ABORTED;	/* XXX Free stuff */
		/* create stdin */
		sprintf (buf,"%010ld", inbuf->used);
		buffer = bufferWrite(buffer, inbuf->read, inbuf->used,
			NSAPIPASS);
	} else { 
		sprintf (buf,"%010ld", 0);
		buffer = bufferWrite(buffer, buf, strlen(buf), NSAPIPASS);
	}


	bufferXmit(buffer, sock, NSAPIPASS);

	bufferReset(buffer, NSAPIPASS);

	/*
	** We have to read the 10 byte count for stdout, then stdout,
	** then the 10 byte count for stderr, then stderr
	**
	** bufferPrint (probably mislabled by now) will handle it
	*/

	printstate.state = 0;	/* an initializer */

	while(bufferRead(buffer, sock, NSAPIPASS)) {
		bufferPrint(buffer, &printstate, NSAPIPASS);
	}

	close(sock);

	bufferFree(buffer);
	bufferFree(inbuf);

	return REQ_EXIT;
}

/*
** dcpcgi_path
**
** Fixup the path components
*/

NSAPI_PUBLIC int dcpcgi_path(NSAPIARGS) {

	char *path = pblock_findval("path-info", rq->vars);
	char *uri  = pblock_findval("uri", rq->reqpb);
	char *base = pblock_findval("base", param);
	char *b;
	char *u;
	char *temp;

	if (base == NULL) return REQ_NOACTION;

	b = base;
	u = uri;

	while (*b && *u && *b == *u) b++, u++;


	temp = STRDUP(u);

	if (path != NULL)
		pblock_replace("path-info",temp,rq->vars);
	else
		pblock_nvinsert("path-info",temp,rq->vars);

	pblock_replace("uri",base,rq->reqpb);

	FREE(temp);

	return REQ_PROCEED;

}
